/*
 *  Copyright 2001 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Reglang FSM.
 *
 *  Reglang FSM is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Reglang FSM is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Reglang FSM; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

#ifndef _RLFSM_FSMBASE_CPP
#define _RLFSM_FSMBASE_CPP

#include <string.h>
#include <assert.h>

/* Copy all graph data including transitions. */
template < class State, class Transition >
		FsmGraph<State, Transition>::
		FsmGraph(const FsmGraph &graph)
:
	/* Lists start empty. Will be filled by copy. */
	stateList(),
	misfitList(),

	entryPoints(graph.entryPoints),

	/* Will be filled by copy. */
	finStateSet(),
	misfitAccounting(false)
{
	/* Create the states and record their map in the original state. */
	State *origState = graph.stateList.head;
	while ( origState != 0 ) {
		/* Make the new state. */
		State *newState = new State(*origState);

		/* Add the state to the list.  */
		stateList.append( newState );

		/* Set the mapsTo item of the old state. */
		origState->alg.stateMap = newState;

		/* next state. */
		origState = origState->next;
	}
	
	/* Derefernce all the state maps. */
	State *state = stateList.head;
	while ( state != 0 ) {
		/* Walk the list of out transitions and attach new transitions to their
		 * corresponding new states. */
		TransEl *tel = state->outList.table;
		int ntel = state->outList.tableLength;
		for (int ot = 0; ot < ntel; ot++, tel++ ) {
			/* Is the transition set? */
			if ( tel->value != NULL ) {
				/* Get the trans and the to state. */
				Transition *trans = tel->value;
				State *to = trans->toState->alg.stateMap;

				/* Make a copy of the transition. */
				trans = new Transition(*trans);
				tel->value = trans;

				/* Simulate a real attaching. */
				trans->fromState = 0;
				trans->toState = 0;
				attachStates( state, to, trans, KeyTypeSingle, tel->key );
			}
		}

		/* Walk the list of out ranges and attach new transitions to their
		 * corresponding new states. */
		tel = state->outRange.table;
		ntel = state->outRange.tableLength;
		for ( int outr = 0; outr < ntel; outr+=2, tel+=2 ) {
			if ( tel->value != NULL ) {
				/* Get the trans and the to state. */
				Transition *trans = tel->value;
				State *to = trans->toState->alg.stateMap;

				/* Make a copy of the transition. */
				trans = new Transition(*trans);
				tel[0].value = trans;
				tel[1].value = trans;

				/* Simulate a real attaching. */
				trans->fromState = 0;
				trans->toState = 0;
				attachStates( state, to, trans, KeyTypeRange, tel->key );
			}
		}

		/* Copy the default transition if it is there. */
		if ( state->outDefault != NULL ) {
			/* Get the trans and the to state. */
			Transition *trans = state->outDefault;
			State *to = trans->toState->alg.stateMap;

			/* Copy the transition. */
			trans = new Transition(*trans);
			state->outDefault = trans;

			/* Simulate a real attaching. */
			trans->fromState = 0;
			trans->toState = 0;
			attachStates( state, to, trans, KeyTypeDefault, 0 );
		}

		state = state->next;
	}

	/* Fix the state pointers in the entry points array. */
	EntryEl *eel = entryPoints.table;
	for ( int e = 0; e < entryPoints.tableLength; e++, eel++ ) {
		/* Get the duplicate of the state. */
		eel->value = eel->value->alg.stateMap;

		/* Foreign in transitions must be built up when duping machines so
		 * icrement it here. */
		eel->value->foreignInTrans += 1;
	}

	/* Build the final state set. */
	State **st = graph.finStateSet.table;
	int nst = graph.finStateSet.tableLength;
	for ( int fs = 0; fs < nst; fs++, st++ )
		finStateSet.set((*st)->alg.stateMap);
}

/* Deletes all transition data then deletes each state. */
template < class State, class Transition >
		FsmGraph<State, Transition>::~FsmGraph()
{
	/* Delete all the transitions. */
	State *state = stateList.head;
	while ( state != NULL ) {
		/* Iterate the out transitions, deleting them. */
		FsmOutIterator<State, Transition> outIt(state);
		for ( ; ! outIt.atEnd(); outIt++ )
			delete outIt.trans;

		state = state->next;
	}

	/* Delete all the states. */
	stateList.empty();
}


/* Set a state final. The state has its isFinState set to true and the state is
 * added to the finStateSet. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		setFinState( State *state )
{
	/* Is it already a fin state. */
	if (state->isFinState)
		return;
	
	state->isFinState = true;
	finStateSet.set( state );
}

/* Set a state non-final. The has its isFinState flag set false and the state
 * is removed from the final state set. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		unsetFinState( State *state )
{
	/* Is it already a non-final state? */
	if ( ! state->isFinState )
		return;

	/* When a state looses its final state status it must relinquish all the
	 * properties that are allowed only for final states. */
	state->relinquishFinal();

	state->isFinState = false;
	finStateSet.unSet( state );
}

/* Associate an id with a state. Makes the state a named entry point. The id
 * must not be in use and the state must not already be an entry point. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		setEntry( int id, State *state )
{
	/* Insert the entry and assert that it succeeds. */
	EntryEl *entryEl = entryPoints.insert( id, state );
	assert( entryEl != 0 );

	/* Remember the entry point in the state. */
	int *setEl = state->entryIds.insert( id );
	assert( setEl != 0 );

	if ( misfitAccounting ) {
		/* If the number of foreign in transitions is about to go up to 1 then
		 * take it off the misfit list and put it on the head list. */
		if ( state->foreignInTrans == 0 )
			stateList.append( misfitList.detach( state ) );
	}

	/* Up the foreign in transitions to the state. */
	state->foreignInTrans += 1;
}

/* Remove the association of an id with a state. The state looses it's entry
 * point status. If id is not in use then will assertion fail. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		unsetEntry( int id )
{
	/* Find the state and assert that it is an entry point. */
	EntryEl *entryEl = entryPoints.find( id );
	assert( entryEl != 0 );

	/* Remove the id from the state's entry id set. */
	bool removeRes = entryEl->value->entryIds.remove( id );
	assert( removeRes );

	/* Decrement the entry's count of foreign entries. */
	entryEl->value->foreignInTrans -= 1;

	if ( misfitAccounting ) {
		/* If the number of foreign in transitions just went down to 0 then take
		 * it off the main list and put it on the misfit list. */
		if ( entryEl->value->foreignInTrans == 0 )
			misfitList.append( stateList.detach( entryEl->value ) );
	}

	/* Remove the record from the map. */
	removeRes = entryPoints.remove( entryEl );
	assert( removeRes );
}

/* Change an entry point to a new state. The entry point must be previously
 * set. The new state the entry is assigned to must not be an entry. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		changeEntry( int id, State *state )
{
	/* Find the state and assert that it is an entry point. */
	EntryEl *entryEl = entryPoints.find( id );
	assert( entryEl != 0 );

	/* Remove the id from the old state's entry id set */
	bool res = entryEl->value->entryIds.remove( id );
	assert( res );

	/* Decrement the old state's count of foreign entries. */
	entryEl->value->foreignInTrans -= 1;

	if ( misfitAccounting ) {
		/* If the number of foreign in transitions just went down to 0 then
		 * take it off the main list and put it on the misfit list. */
		if ( entryEl->value->foreignInTrans == 0 )
			misfitList.append( stateList.detach( entryEl->value ) );
	}

	/* Set the id in the new state's entry id set. */
	int *setEl = state->entryIds.insert( id );
	assert( setEl != 0 );

	if ( misfitAccounting ) {
		/* If the number of foreign in transitions is about to go up to 1 then
		 * take it off the misfit list and put it on the head list. */
		if ( state->foreignInTrans == 0 )
			stateList.append( misfitList.detach( state ) );
	}

	/* Up the foreign in transitions of the new state. */
	state->foreignInTrans += 1;

	/* Set the new state in the entry point array. */
	entryEl->value = state;
}


/* Find a state by an entry id. If the id does not represent and entry point
 * then returns null. */
template < class State, class Transition >
		State *FsmGraph<State, Transition>::
		findEntry( int id ) const
{
	State *retVal = 0;

	EntryEl *entry = entryPoints.find( id );
	if ( entry != 0 )
		retVal = entry->value;

	return retVal;
}

/* Set the priority of starting transitions. Isolates the start state so it has
 * no other entry points, then sets the priorities of all the transitions out
 * of the start state. If the start state is final, then the outPrior of the
 * start state is also set. The idea is that a machine that accepts the null
 * string can still specify the starting trans prior for when it accepts the
 * null word. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		startFsmPrior( int prior )
{
	/* Make sure the start state has no other entry points. */
	isolateStartState();

	/* Get the start state. */
	State *startState = findEntry( 0 );

	/* Walk all transitions out of the start state. */
	FsmOutIterator<State, Transition> outIt( startState );
	for ( ; ! outIt.atEnd(); outIt++ )
		outIt.trans->priority = prior;

	/* If the new start state is final then set the out priority. This follows
	 * the same convention as setting start funcs in the out funcs of a final
	 * start state. */
	if ( startState->isFinState ) {
		startState->isOutPriorSet = true;
		startState->outPriority = prior;
	}
}

/* Set the priority of all transitions in a graph. Walks all transition lists
 * and all def transitions. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		allTransPrior( int prior )
{
	/* Walk the out list of all states. */
	State *state = stateList.head;
	while ( state != NULL ) {
		/* Walk the out list of the state. */
		FsmOutIterator<State, Transition> outIt( state );
		for ( ; ! outIt.atEnd() ; outIt++ ) 
			outIt.trans->priority = prior;

		state = state->next;
	}
}

/* Set the priority of all transitions that go into a final state. Note that if
 * any entry states are final, we will not be setting the priority of any
 * transitions that may go into those states in the future. The graph does not
 * support pending in transitions in the same way pending out transitios are
 * supported. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		finFsmPrior( int prior )
{
	/* Walk all final states. */
	State **st = finStateSet.table;
	int nst = finStateSet.tableLength;
	for (int i = 0; i < nst; i++, st++) {
		/* Walk all in transitions of the final state. */
		FsmInIterator<State, Transition> inIt( *st );
		for ( ; ! inIt.atEnd(); inIt++ )
			inIt.trans->priority = prior;
	}
}

/* Set the priority of any future out transitions that may be made going out of
 * this state machine. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		leaveFsmPrior( int prior )
{
	/* Walk all final states. */
	State **st = finStateSet.table;
	int nst = finStateSet.tableLength;

	for (int i = 0; i < nst; i++, st++) {
		(*st)->isOutPriorSet = true;
		(*st)->outPriority = prior;
	}
}


/* Set functions to execute on starting transitions. Isolates the start state
 * so it has no other entry points, then adds to the transition functions
 * of all the transitions out of the start state. If the start state is final,
 * then the func is also added to the start state's out func list. The idea is
 * that a machine that accepts the null string can execute a start func when it
 * matches the null word, which can only be done when leaving the start/final
 * state. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		startFsmFunc( int func, int transOrder )
{
	/* Make sure the start state has no other entry points. */
	isolateStartState();

	/* Get the start state. */
	State *startState = findEntry( 0 );

	/* Walk the start state's transitions, setting functions. */
	FsmOutIterator<State, Transition> outIt( startState );
	for ( ; ! outIt.atEnd(); outIt++ ) 
		outIt.trans->setFunction( func, transOrder );

	/* If start state is final then insert on the out func. This means that you
	 * can have start and leaving funcs on a machine that recognizes the null
	 * word. */
	if ( startState->isFinState )
		startState->outTransFuncTable.insertMulti( transOrder, func );
}

/* Set functions to execute on all transitions. Walks the out lists of all
 * states. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		allTransFunc( int func, int transOrder )
{
	/* Walk all states. */
	State *state = stateList.head;
	while ( state != NULL ) {
		/* Walk the out list of the state. */
		FsmOutIterator<State, Transition> outIt( state );
		for ( ; ! outIt.atEnd(); outIt++ )
			outIt.trans->setFunction( func, transOrder );

		state = state->next;
	}
}

/* Specify functions to execute upon entering final states. If the start state
 * is final we can't really specify a function to execute upon entering that
 * final state the first time. So function really means whenever entering a
 * final state from within the same fsm. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		finFsmFunc( int func, int transOrder )
{
	/* Walk all final states. */
	State **st = finStateSet.table;
	int nst = finStateSet.tableLength;
	for (int i = 0; i < nst; i++, st++) {
		/* Walk the final state's in list. */
		FsmInIterator<State, Transition> inIt( *st );
		for ( ; ! inIt.atEnd(); inIt++ )
			inIt.trans->setFunction( func, transOrder );
	}
}

/* Add functions to any future out transitions that may be made going out of
 * this state machine. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		leaveFsmFunc( int func, int transOrder )
{
	State **st = finStateSet.table;
	int nst = finStateSet.tableLength;
	for (int i = 0; i < nst; i++, st++)
		(*st)->outTransFuncTable.insertMulti( transOrder, func );
}

/* Shift the function ordering of the start transitions to start
 * at fromOrder and increase in units of 1. Useful before staring.
 * Returns the maximum number of order numbers used. */
template < class State, class Transition >
		int FsmGraph<State, Transition>::
		shiftStartFuncOrder( int fromOrder )
{
	int maxUsed = 0;

	/* Walk the start state's transitions, shifting function ordering. */
	FsmOutIterator<State, Transition> outIt( findEntry(0) );
	for ( ; ! outIt.atEnd(); outIt++ ) {
		Transition *trans = outIt.trans;

		/* Walk the function table for the transition. */
		typename Transition::TransFuncEl *tfel =
				trans->transFuncTable.table;

		/* Set the keys to increasing values starting at fromOrder */
		int ntfel = trans->transFuncTable.tableLength;
		int curFromOrder = fromOrder;
		for ( int j = 0; j < ntfel; j++, tfel++ )
			tfel->key = curFromOrder++;
	
		/* Keep track of the max number of orders used. */
		if ( curFromOrder - fromOrder > maxUsed )
			maxUsed = curFromOrder - fromOrder;
	}
	
	return maxUsed;
}

/* Clear all functions tables out of the start state. First makes sure that
 * the start state has no other entry points other than its start stateness. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		clearStartFsmFunc()
{
	/* Make sure the start state has no other entry points. */
	isolateStartState();

	/* Get the start state. */
	State *startState = findEntry( 0 );

	/* Walk the start state's transitions, clearing functions. */
	FsmOutIterator<State, Transition> outIt( startState );
	for ( ; ! outIt.atEnd(); outIt++ )
		outIt.trans->transFuncTable.empty();

	/* If start state is final then clear on the out func. Leaving the start state
	 * via an out transition is considered starting the fsm. */
	if ( startState->isFinState )
		startState->outTransFuncTable.empty();
}

/* Empty all transition functions tables. Does not empty state outFunc tables. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		clearAllTransFunc()
{
	/* Walk all states. */
	State *state = stateList.head;
	while ( state != NULL ) {
		/* Walk the state's outList. */
		FsmOutIterator<State, Transition> outIt( state );
		for ( ; ! outIt.atEnd(); outIt++ )
			outIt.trans->transFuncTable.empty();

		state = state->next;
	}
}

/* Empty all transition functions going into a final state. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		clearFinFsmFunc()
{
	/* Walk all final states. */
	State **st = finStateSet.table;
	int nst = finStateSet.tableLength;
	for (int i = 0; i < nst; i++, st++) {
		/* Walk the final state's in list. */
		FsmInIterator<State, Transition> inIt( *st );
		for ( ; ! inIt.atEnd(); inIt++ )
			inIt.trans->transFuncTable.empty();
	}
}

/* Clear pending out functions from the outFunc lists of final states. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		clearLeaveFsmFunc()
{
	State **st = finStateSet.table;
	int nst = finStateSet.tableLength;
	for (int i = 0; i < nst; i++, st++)
		(*st)->outTransFuncTable.empty();
}

/* Clear pending out priorities from final states. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		clearLeaveFsmPrior( )
{
	State **st = finStateSet.table;
	int nst = finStateSet.tableLength;

	for (int i = 0; i < nst; i++, st++) {
		(*st)->isOutPriorSet = false;
		(*st)->outPriority = 0;
	}
}

/* Remove all transition data. Remove funcs and outfuncs, zero priorities and
 * remove out priorities. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		clearAllTransData()
{
	State *state = stateList.head;
	while ( state ) {
		/* Clear out data. */
		state->outTransFuncTable.empty();
		state->isOutPriorSet = false;
		state->outPriority = 0;

		/* Clear transition data from the out transitions. */
		FsmOutIterator<State, Transition> outIt( state );
		for ( ; ! outIt.atEnd(); outIt++ ) {
			outIt.trans->transFuncTable.empty();
			outIt.trans->priority = 0;
		}

		state = state->next;
	}
}


/* Zeros out the function ordering keys. This may be called before minimization
 * when it is known that no more fsm operations are going to be done.  This
 * will achieve greater reduction as states will not be separated on the basis
 * of function ordering. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		nullFunctionKeys( )
{
	/* For each state... */
	State *state = stateList.head;
	while ( state != NULL ) {
		/* Walk the transitions for the state. */
		FsmOutIterator<State, Transition> outIt( state );
		for ( ; ! outIt.atEnd(); outIt ++ ) {
			/* Walk the function table for the transition. */
			typename Transition::TransFuncEl *tfel = outIt.trans->transFuncTable.table;
			int ntfel = outIt.trans->transFuncTable.tableLength;
			for ( int j = 0; j < ntfel; j++, tfel++ )
				tfel->key = 0;
		}

		/* Null the function keys of the out transtions. */
		typename Transition::TransFuncEl *tfel = state->outTransFuncTable.table;
		int ntfel = state->outTransFuncTable.tableLength;
		for ( int j = 0; j < ntfel; j++, tfel++ )
			tfel->key = 0;

		state = state->next;
	}
}

/* Mark all states reachable from state. Traverses transitions forward. Used
 * for removing states that have no path into them. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		markReachableFromHere( State *state )
{
	/* Base case: return; */
	if ( state->isMarked )
		return;
	
	/* Set this state as processed. We are going to visit all states that this
	 * state has a transition to. */
	state->isMarked = true;

	/* Recurse on all out transitions. */
	FsmOutIterator<State, Transition> outIt( state );
	for ( ; ! outIt.atEnd(); outIt++ )
		markReachableFromHere( outIt.trans->toState );
}

/* Mark all states reachable from state. Traverse transitions backwards. Used
 * for removing dead end paths in graphs. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		markReachableFromHereReverse( State *state )
{
	/* Base case: return; */
	if ( state->isMarked )
		return;
	
	/* Set this state as processed. We are going to visit all states with
	 * transitions into this state. */
	state->isMarked = true;

	/* Recurse on all items in transitions. */
	FsmInIterator<State, Transition> inIt( state );
	for ( ; ! inIt.atEnd(); inIt++ )
		markReachableFromHereReverse( inIt.trans->fromState );
}


/* Tests the integrity of the transition lists and the fromStates. */
template < class State, class Transition >
		void FsmGraph<State, Transition>::
		verifyIntegrity()
{
	State *state = stateList.head;
	while ( state != NULL ) {
		/* Walk the out transitions and assert fromState is correct. */
		FsmOutIterator<State, Transition> outIt( state );
		for ( ; ! outIt.atEnd(); outIt++ )
			assert( outIt.trans->fromState == state );

		/* Walk the inlist and assert toState is correct. */
		FsmInIterator<State, Transition> inIt( state );
		for ( ; ! inIt.atEnd(); inIt++ )
			assert( inIt.trans->toState == state );

		state = state->next;
	}
}

#endif /* _RLFSM_FSMBASE_CPP */
