/*
 *  Copyright 2001 Adrian Thurston <adriant@ragel.ca>
 */

/*  This file is part of Ragel.
 *
 *  Ragel is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 * 
 *  Ragel is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with Ragel; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

%{

#include <stdlib.h>
#include "rl.h"
#include "fsm.h"
#include "parsetree.h"

/* Generate a fsm from a graph dict. In main.cc. */
void GenerateFsm( char *fsmName, ParseData *parseData );

/* Ids to use for builtin functions. */
int builtinNextId = 1;

/* The parse data. The parser collects the things that it parses in
 * the data structures in here. */
ParseData *parseData;

/* Minimization after an operator. */
void AfterOpMinimize(Fsm *fsm);

%}

%union {
	char *data;
	char chr;
	int integer;
	TermNode *term;
	FactorWithRepNode *factorWithRep;
	FactorWithAugNode *factorWithAug;
	FactorNode *factor;
	ExpressionNode *expression;
	int pAug;
	int funcId;
	AugType augType;
	Fsm *fsm;
	RegExpSet regExpSet;
}

%token <data> PP_Literal
%token <data> PP_Integer
%token <data> PP_Word
%token <data> PP_Comment
%token <data> PP_Character
%token <data> PP_Whitespace
%token <data> PP_Section

%type <data> NonFsmAny
%type <data> FsmName

%token <data> RL_Integer
%token <data> RL_Hex
%token <data> RL_Word
%token <data> RL_Literal
%token <data> RL_DoubleDot

%token <chr> RL_ReLiteralSlash
%token <chr> RL_ReLiteralChar
%token <chr> RL_ReLiteralOpen
%token <chr> RL_ReLiteralOpenNeg
%token <chr> RL_ReLiteralClose
%token <chr> RL_ReLiteralDot
%token <chr> RL_ReLiteralStar
%token <regExpSet> RL_ReLiteralSet

%token <data> RL_OrLiteral
%token <data> RL_Builtin
%token <data> RL_Data
%token <data> RL_Func
%token <data> RL_Init
%token <data> RL_Clear
%token <data> '*' '?' '+' '!' '(' ')' ';' '=' ':' '%' '$' '-' '|' '&' '.' '>'

%type <augType> AugType
%type <pAug> PriorityAug
%type <funcId> FunctionCall
%type <expression> Expression 
%type <term> Term
%type <factorWithAug> FactorWithAug 
%type <factorWithRep> FactorWithRep 
%type <factor> Factor
%type <integer> RangeLit

%type <fsm> RegularExpr
%type <fsm> RegularExprItem
%type <fsm> RegularExprChar
%type <fsm> RegularExprOrData
%type <fsm> RegularExprOrChar

%%

/* Input is any number of input sections. An empty file is accepted. */
input: Sections ;
Sections: 		
	Sections SectionItem { } |
	{ }
	;

/* A section item consists of either a single item from the input file (could
 * be a comment, string, identifier) which is passed through to the output
 * file as is or could be a fsm specification in which case it is parsed. */
SectionItem:
	NonFsmAny {
		// Write the data out to the output stream. */
		*outputStream << $1;
	} |
	FsmSpec
	;

/* The types of tokens that can appear outside of a fsm specifications
 * They are the alternate to a fsm specification in the SetionItem. */
NonFsmAny:		PP_Literal |
				PP_Integer |
				PP_Word |
				PP_Comment |
				PP_Character |
				PP_Whitespace
				;


/* Fsm Specification. Fsms are deliminated by the %% tokens 
 * and require a name for the fsm immediately following the open %% 
 * deliminator. */
FsmSpec:
	PP_Section FsmName StatementList PP_Section {
		// Process the fsm.
		GenerateFsm( $2, parseData );
		delete parseData;
	}
	;

/* The Name of the fsm. Use this production to initialize the globals
 * used during the parse of the fsm. */
FsmName:
	RL_Word {
		// Take this opportunity to new up everything
		// necessary to process the fsm
		parseData = new ParseData();
		$$ = $1;
	}
	;


/* A NonEmpty list of statements in a fsm. */
StatementList:
	StatementList Statement |
	Statement
	;

/* The differnt types of statements in a fsm spec. */
Statement:	
	RLAssignment |
	DataSpec |
	FuncSpec |
	InitSpec;

/* Allow the user to give some code to insert into the init function. */
InitSpec:
	RL_Init RL_Builtin {
	// Put the code on the init code list. 
		parseData->initCodeList.append( new StringListEl( $2 ) );
	} ;

/* Allow the user to give some declarations to put into the fsm structure. */
DataSpec:	
	RL_Data RL_Builtin ';' {
		// Put the spec for the data on the list. Note that we use
		// a func list el but it is data.
		parseData->dataList.append( new StringListEl( $2 ) );
	} ;

/* Allow the user to create a named fsm function that can be referenced when
 * building a machine. */
FuncSpec:	
	RL_Func RL_Word RL_Builtin {
		if ( parseData->funcDict.insert( $2,
						parseData->nextFuncNum ) ) {
			// Add the function to the list of functions.
			parseData->funcList.append( new StringListEl( $3 ) );

			// Used up a function number.
			parseData->nextFuncNum += 1;
		}
		else {
			fprintf( stderr, "%s:%i: warning: func \"%s\" "
				"already defined\n", inputFile, rlCurLine, $2);
		}
	} ;

/* An assignement statement. Makes an fsm and assigns it to a variable. */
RLAssignment:
	RL_Word '=' Expression ';' {
		if ( ! parseData->graphDict.insert($1, $3) ) {
			fprintf( stderr, "%s:%i: warning: variable \"%s\" "
   		         "already defined\n", inputFile, rlCurLine, $1);
		}
		parseData->machineGiven = true;
	}
	;

/* Top level production in the parse of a fsm. The lowest precedence
 * is the '|' (or), '&' (intersection), and '-' (subtraction) operators. */
Expression:
	Expression '|' Term {
		$$ = new ExpressionNode( $1, $3, ExpressionNode::Or );
	} |
	Expression '&' Term {
		$$ = new ExpressionNode( $1, $3, ExpressionNode::Intersect );
	} |
	Expression '-' Term {
		$$ = new ExpressionNode( $1, $3, ExpressionNode::Subtract );
	} |
	Term {
		$$ = new ExpressionNode( $1 );
	}
	;

/* The second group up in precedence. The concatenation operators. The special
 * '.^' operator forces the leaving fsm flag to the Concat operation to be 
 * false. The result is that the concatenation ignores out functions and out
 * priorities. */
Term:
	Term FactorWithAug {
		$$ = new TermNode( true, $1, $2 );
	} |
	Term '.' FactorWithAug {
		$$ = new TermNode( true, $1, $3 );
	} |
	Term '.' '^' FactorWithAug {
		$$ = new TermNode( false, $1, $4 );
	} |
	FactorWithAug {
		$$ = new TermNode( $1 );
	}
	;

/* Third group up in precenedce. Allow users to add function calls and to change
 * the priorities of transitions. */
FactorWithAug:
	FactorWithAug AugType FunctionCall {
		$$ = $1;
		$$->funcs.append( ParserFunc( $3, 0, $2 ) );
	} |
	FactorWithAug AugType PriorityAug {
		$$ = $1;
		$$->priorityAugs.append( PriorityAug( $3, $2 ) );
	} |
	FactorWithAug '%' '!' {
		$$ = $1;
		$$->priorityAugs.append( PriorityAug( 0, clearLeave ) );
	} |
	FactorWithRep {
		$$ = new FactorWithAugNode( $1 );
	}
	;

/* Priority change specs for FactorWithAug. */
PriorityAug:
	'-' RL_Integer {
		$$ = -1 * atoi( $2 );
	} |
	'+' RL_Integer {
		$$ = atoi( $2 );
	} |
	RL_Integer {
		$$ = atoi( $1 );
	}
	;
		
/* Differnt places to change the priorities or to add function calls. 
 * Used by FactorWithAug for both types of augmentation. */
AugType:
	':' { $$ = fin; } |
	'%' { $$ = leave; } |
	'$' { $$ = all; } |
	'>' { $$ = start; }
	;

/* Different types of function call specs. A RL_Word is reference to a func
 * given by the user as a statement in the fsm specification. A function
 * can also be specified immediately. */
FunctionCall:
	RL_Word {
		// Set the name in the funcDict.
		char *funcName = $1;
		int func;
		FuncDictEl *fdel = parseData->funcDict.find( funcName );
		if ( fdel != 0 ) {
			func = fdel->value;
		}
		else {
			fprintf( stderr, "%s:%i: warning: action lookup"
				"of \"%s\" failed. using null action\n",
				inputFile, rlCurLine, $1 );
			func = 0;
		}
		// Pass up the function name;
		$$ = func;
	} |
	RL_Builtin {
		// Add the function to the list of functions.
		parseData->funcList.append( new StringListEl( $1 ) );

		// Pass up the builtin name.
		$$ = parseData->nextFuncNum;
		parseData->nextFuncNum += 1;
	} |
	RL_Clear {
		// This is a clear action, Don't set the action, clear
		// the actions instead.
		$$ = FUNC_NUM_CLEAR;
	}
	;

/* The fourth level of precedence. These are the unary operators that allow
 * for repetition and negating. */
FactorWithRep:
	'*' FactorWithRep {
		$$ = new FactorWithRepNode( 
					true, $2, FactorWithRepNode::Star );
	} |
	'*' '^' FactorWithRep {
		$$ = new FactorWithRepNode( 
					false, $3, FactorWithRepNode::Star );
	} |
	'?' FactorWithRep {
		$$ = new FactorWithRepNode( 
					true, $2, FactorWithRepNode::Optional );
	} |
	'+' FactorWithRep {
		$$ = new FactorWithRepNode( 
					true, $2, FactorWithRepNode::Plus );
	} |
	'+' '^' FactorWithRep {
		$$ = new FactorWithRepNode( 
					false, $3, FactorWithRepNode::Plus );
	} |
	'!' FactorWithRep {
		$$ = new FactorWithRepNode(
					true, $2, FactorWithRepNode::Negate );
	} |
	Factor {
		$$ = new FactorWithRepNode( $1 );
	}
	;

/* The highest level in precedence. Atomic machines such as references to other
 * machines, literal machines, regular expressions or Expressions in side of
 * parenthesis. */
Factor:
	RL_Literal {
		if ( *($1) == 0 )
			$$ = new FactorNode( Fsm::nullFsm() );
		else
			$$ = new FactorNode( Fsm::concatFsm( $1, strlen($1) ) );
	} |

	RL_OrLiteral {
		if ( *($1) == 0 )
			$$ = new FactorNode( Fsm::nullFsm() );
		else {
			char seen[256];
			memset(seen, 0, 256);
			for ( char *p = $1; *p != 0; ) {
				if ( seen[(unsigned char)*p] ) {
					// Shift over and toast *p.
					memmove(p, p+1, strlen(p+1)+1);
				}
				else {
					// next p, *p is good.
					seen[(unsigned char)*p] = 1;
					p += 1;
				}
			}
			$$ = new FactorNode( Fsm::orFsm( $1, strlen($1) ) );
		}
	} |

	RL_ReLiteralSlash RegularExpr RL_ReLiteralSlash {
		$$ = new FactorNode( $2 );
	} |

	RL_Integer {
		int intFsm = atoi( $1 );
		// Restrict to byte size, cannot not be negative.
		if ( intFsm >= 256 ) {
			fprintf( stderr, "%s:%i: warning: overflow in byte constant\n",
				inputFile, rlCurLine );
		}
		$$ = new FactorNode( Fsm::concatFsm( intFsm ) );
	} |

	RL_Hex {
		int hexFsm = strtoul( $1, 0, 16 );
		// Restrict to byte size, cannot not be negative.
		if ( hexFsm >= 256 ) {
			fprintf( stderr, "%s:%i: warning: overflow in byte constant\n",
				inputFile, rlCurLine );
		}
		$$ = new FactorNode( Fsm::concatFsm( hexFsm ) );
	} |

	RL_Word {
		GraphDictNode *gdNode = parseData->graphDict.find( $1 );
		if ( gdNode != NULL ) {
			$$ = new FactorNode( gdNode->value, FactorNode::LookupExpression );
		}
		else {
			fprintf( stderr, "%s:%i: warning: graph lookup"
				"of \"%s\" failed, using null fsm\n",
				inputFile, rlCurLine, $1 );
			$$ = new FactorNode( Fsm::nullFsm() );
		}
	} |

	RangeLit RL_DoubleDot RangeLit {
		if ( $1 > $3 ) {
			fprintf( stderr, "%s:%i: warning: range gives null fsm\n",
					inputFile, rlCurLine );
		}
		$$ = new FactorNode( $1, $3 );
	} |

	'(' Expression ')' {
		$$ = new FactorNode( $2, FactorNode::Expression );
	}
	;

/* Parser for regular expression fsms. Any number of expression items which
 * generally gives a machine one character long or one character long stared. */
RegularExpr:
	RegularExpr RegularExprItem {
		// concat together, return.
		if ( $1 == 0 ) {
			$$ = $2;
		}
		else {
			$1->concatOp( $2, false );
			$$ = $1;
		}
	} |
	/* Nothing */ {
		$$ = 0;
	}
	;

/* RegularExprItems can be a character spec with an optional staring of the char. */
RegularExprItem:
	RegularExprChar RL_ReLiteralStar {
		$$ = $1;
		$$->starOp(false);
	} |
	RegularExprChar {
		$$ = $1;
	}
	;

/* A character spec can be a set of characters inside of square parenthesis,
 * a dot specifying any character or some explicitly stated character. */
RegularExprChar:
	RL_ReLiteralOpen RegularExprOrData RL_ReLiteralClose {
		$$ = $2;
		$$->minimizePartition2();
	} |
	RL_ReLiteralOpenNeg RegularExprOrData RL_ReLiteralClose {
		$$ = Fsm::dotFsm();
		$2->minimizePartition2();
		$$->subtractOp( $2 );
		$$->minimizePartition2();
	} |
	RL_ReLiteralDot {
		$$ = Fsm::dotFsm();
	} |
	RL_ReLiteralChar {
		char str[2] = { $1, 0 };
		$$ = Fsm::concatFsm( str, strlen(str) );
	}
	;

/* The data inside of a [] expression in a regular expression. Accepts any
 * number of characters or ranges. */
RegularExprOrData:
	RegularExprOrData RegularExprOrChar {
		if ( $1 == 0 ) {
			$$ = $2;
		}
		else {
			$1->orOp( $2 );
			$$ = $1;
		}
	} | 
	/* Nothing */ {
		$$ = 0;
	}
	;


/* A single character inside of an or expression. Can either be a character
 * or a set of characters. */
RegularExprOrChar:
	RL_ReLiteralChar {
		char str[2] = { $1, 0 };
		$$ = Fsm::concatFsm( str, strlen(str) );
	} |
	RL_ReLiteralSet {
		int lower = (unsigned char) $1.lower;
		int upper = (unsigned char) $1.upper;
		$$ = Fsm::rangeFsm( lower, upper );
	}
	;

RangeLit:
	RL_Literal {
		if ( strlen($1) != 1 ) {
			fprintf( stderr, "%s:%i: warning: literal used in range is not "
					"of length 1, using 0x%x\n", inputFile, rlCurLine,
					(unsigned char) *($1) );
		}
		$$ = (unsigned char) *($1);
	} |

	RL_Integer {
		int intFsm = atoi( $1 );
		// Restrict to byte size, cannot not be negative.
		if ( intFsm >= 256 ) {
			fprintf( stderr, "%s:%i: warning: overflow in byte constant\n",
				inputFile, rlCurLine );
		}
		$$ = intFsm;
	} |

	RL_Hex {
		int hexFsm = strtoul( $1, 0, 16 );
		// Restrict to byte size, cannot not be negative.
		if ( hexFsm >= 256 ) {
			fprintf( stderr, "%s:%i: warning: overflow in byte constant\n",
				inputFile, rlCurLine );
		}
		$$ = hexFsm;
	}
	;

	
%%

void rlerror(char *err)
{
	fprintf(stderr, "%s:%i: %s\n", inputFile, rlCurLine, err);
	exit(1);
}
