/*
 * A mini C-like language scanner.
 */

#include <stdio.h>

#define IDENT_BUFLEN 256

%% Fsm
	# The data that is to go into the fsm structure.
	struct {
		char identBuf[IDENT_BUFLEN+1];
		int identLen;
		int curLine;
	};

	# Initialization code that will go into the ClangInit routine.
	init {
		fsm->identLen = 0;
		fsm->curLine = 1;
	}

	# Function to buffer a character.
	func bufChar {
		if ( fsm->identLen < IDENT_BUFLEN ) {
			fsm->identBuf[fsm->identLen] = *p;
			fsm->identLen += 1;
		}
	}

	# Function to clear the buffer.
	func clearBuf {
		fsm->identLen = 0;
	}

	# Functions to dump tokens as they are matched.
	func ident {
		fsm->identBuf[fsm->identLen] = 0;
		printf("ident(%i): %s\n", fsm->curLine, fsm->identBuf);
	}
	func literal {
		fsm->identBuf[fsm->identLen] = 0;
		printf("literal(%i): \"%s\"\n", fsm->curLine, fsm->identBuf);
	}
	func float {
		fsm->identBuf[fsm->identLen] = 0;
		printf("float(%i): %s\n", fsm->curLine, fsm->identBuf);
	}
	func int {
		fsm->identBuf[fsm->identLen] = 0;
		printf("int(%i): %s\n", fsm->curLine, fsm->identBuf);
	}
	func hex {
		fsm->identBuf[fsm->identLen] = 0;
		printf("hex(%i): %s\n", fsm->curLine, fsm->identBuf);
	}
	func symbol {
		fsm->identBuf[fsm->identLen] = 0;
		printf("symbol(%i): %s\n", fsm->curLine, fsm->identBuf);
	}

	# Alpha numberic characters or underscore.
	alnumu = alnum | '_';

	# Alpha charactres or underscore.
	alphau = alpha | '_';

	# Symbols. Upon entering clear the buffer. On all transitions
	# buffer a character. Upon leaving dump the symbol.
	symbol = ( punct - [_] ) >clearBuf $bufChar %symbol;

	# Identifier. Upon entering clear the buffer. On all transitions
	# buffer a character. Upon leaving, dump the identifier.
	ident = (alphau . *alnumu) >clearBuf $bufChar %ident;

	# Match single characters inside literal strings. Or match 
	# an escape sequence. Buffers the charater matched.
	sliteralChar = 
			( extend - ['\\] ) :bufChar |
			( '\\' . extend :bufChar );
	dliteralChar = 
			( extend - ["\\] ) :bufChar |
			( '\\' . extend :bufChar );

	# Single quote and double quota literals. At the start clear
	# the buffer. Upon leaving dump the literal.
	sliteral = ('\'' :clearBuf . *sliteralChar . '\'' ) %literal;
	dliteral = ('"' :clearBuf . *dliteralChar . '"' ) %literal;
	literal = sliteral | dliteral;

	# Whitespace is standard ws, newlines and control codes.
	whitespace = 0x0..0x20 | 0x7f..0xff;

	# Describe both c style comments and c++ style comments. The
	# priority bump on tne terminator of the comments brings us
	# out of the *extend which matches everything.
	ccComment = '//' . *extend . '\n' :1;
	cComment = '/*' . *extend . '*/' :1;

	# Match an integer. We don't bother clearing the buf or filling it.
	# The float machine overlaps with int and it will do it.
	int = +digit %int;

	# Match a float. Upon entering the machine clear the buf, buffer
	# characters on every trans and dump the float upon leaving.
	float =  ( +digit . '.' . +digit ) >clearBuf $bufChar %float;

	# Match a hex. Upon entering clear the buf, buffer characters
	# on every trans and dump the hex on leaving transitions.
	hex = ('0x' . +xdigit) >clearBuf $bufChar %hex;

	# Or together all the lanuage elements.
	fin = ( ccComment |
		cComment |
		symbol |
		ident |
		literal |
		whitespace |
		int |
		float |
		hex );

	# Star the language elements. It is critical in this type of application
	# that we decrease the priority of out transitions before doing so. This
	# is so that when we see 'aa' we stay in the fin machine to match an ident
	# of length two and not wrap around to the front to match two idents of 
	# length one.
	clang_main = *( fin % -1 );

	# This machine matches everything, taking note of newlines.
	newline = *( extend | '\n' : { fsm->curLine += 1; } );

	# The final fsm is the lexer unioned with the newline machine which
	# will count lines for us. The two machines will run concurrently and
	# independently provided we reset the priority of all transitions in 
	# the lexer fsm so that the it does not interfere with the line counter.
	main = 
		clang_main $0 %! | 
		newline;
%%

